from typing import Optional, Tuple


class LRCProvider:
    """
    Base class for all of the synced (LRC format) lyrics providers.
    """

    def __init__(self, session) -> None:
        self.session = session
        self.session.headers.update(
            {
                "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36"
            }
        )

    async def get_lrc_by_id(self, track_id: str) -> Optional[str]:
        """
        Returns the synced lyrics of the song in [LRC](https://en.wikipedia.org/wiki/LRC_(file_format)) format if found.

        ### Arguments
        - track_id: The ID of the track defined in the provider database. e.g. Spotify/Deezer track ID
        """
        raise NotImplementedError

    async def get_lrc(self, search_term: str, duration: int = -1, max_deviation: int = 2000) -> Tuple[Optional[str], int, int]:
        """
        Returns the synced lyrics of the song in [LRC](https://en.wikipedia.org/wiki/LRC_(file_format)) format if found.
        (lyrics, ratio, duration_offset)
        """
        raise NotImplementedError



import argparse
import logging
import asyncio
from syncedlyrics_aio import search


def cli_handler():
    """
    Console entry point handler function.
    This parses the CLI arguments passed to `syncedlyrics -OPTIONS` command
    """
    parser = argparse.ArgumentParser(
        description="Search for an LRC format (synchronized lyrics) of a music"
    )
    parser.add_argument("search_term", help="The search term to find the track.")
    parser.add_argument(
        "-o", "--output", help="Path to save '.lrc' lyrics", default="{search_term}.lrc"
    )
    parser.add_argument(
        "--allow-plain",
        help="Return a plain text (not synced) lyrics if not LRC was found",
        action="store_true",
    )
    parser.add_argument(
        "-v", "--verbose", help="Use this flag to show the logs", action="store_true"
    )
    parser.add_argument(
        "-p", "--providers", help="Lrc providers", default=""
    )
    parser.add_argument(
        "-d", "--duration", help="The duration of track in ms. Set negative if unknow", type=int, default=-1
    )
    parser.add_argument(
        "-m", "--max_deviation", help="Max deviation for a subtitle length in ms, enable if duration is positive", type=int, default=2000
    )
    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if args.verbose else logging.INFO)

    loop = asyncio.get_event_loop()
    lrc = loop.run_until_complete(search(args.search_term, args.allow_plain, args.output, args.providers.split(), args.duration, args.max_deviation))
    if lrc:
        print(lrc)



"""Musixmatch LRC provider"""

import json
from typing import List, Optional, Tuple
from .base import LRCProvider
import time
import os
from rapidfuzz.fuzz import token_sort_ratio

# Inspired from https://github.com/Marekkon5/onetagger/blob/0654131188c4df2b4b171ded7cdb927a4369746e/crates/onetagger-platforms/src/musixmatch.rs
# Huge part converted from Rust to Py by ChatGPT :)


class Musixmatch(LRCProvider):
    """Musixmatch provider class"""

    ROOT_URL = "https://apic-desktop.musixmatch.com/ws/1.1/"

    def __init__(self, session) -> None:
        super().__init__(session)
        self.token = None
        self.session.headers.update(
            {
                "authority": "apic-desktop.musixmatch.com",
                "cookie": "AWSELBCORS=0; AWSELB=0",
            }
        )

    async def _get(self, action: str, query: List[tuple]):
        if action != "token.get" and self.token is None:
            await self._get_token()
        query.append(("app_id", "web-desktop-app-v1.0"))
        if self.token is not None:
            query.append(("usertoken", self.token))
        t = str(int(time.time() * 1000))
        query.append(("t", t))
        url = self.ROOT_URL + action
        response = await self.session.get(url, params=query)
        return response

    async def _get_token(self):
        # Check if token is cached and not expired
        token_path = os.path.join(".syncedlyrics", "musixmatch_token.json")
        current_time = int(time.time())
        if os.path.exists(token_path):
            with open(token_path, "r") as token_file:
                cached_token_data = json.load(token_file)
            cached_token = cached_token_data.get("token")
            expiration_time = cached_token_data.get("expiration_time")
            if cached_token and expiration_time and current_time < expiration_time:
                self.token = cached_token
                return
        # Token not cached or expired, fetch a new token
        r = await self._get("token.get", [("user_language", "en")])
        d = json.loads(await r.text())
        if d["message"]["header"]["status_code"] == 401:
            time.sleep(10)
            return await self._get_token()
        new_token = d["message"]["body"]["user_token"]
        expiration_time = current_time + 600  # 10 minutes expiration
        # Cache the new token
        self.token = new_token
        token_data = {"token": new_token, "expiration_time": expiration_time}
        os.makedirs(".syncedlyrics", exist_ok=True)
        with open(token_path, "w") as token_file:
            json.dump(token_data, token_file)

    async def get_lrc_by_id(self, track_id: str) -> Optional[str]:
        r = await self._get(
            "track.subtitle.get", [("track_id", track_id), ("subtitle_format", "lrc")]
        )
        if not r.ok:
            return
        _json = json.loads(await r.text())
        body = _json["message"]["body"]

        if not body:
            return
        return body["subtitle"]["subtitle_body"]

    async def get_lrc(self, search_term: str, duration: int = -1, max_deviation: int = 2000) -> Tuple[Optional[str], int]:
        """Duration NotImplemented"""
        r = await self._get(
            "track.search",
            [
                ("q", search_term),
                ("page_size", "5"),
                ("page", "1"),
                ("s_track_rating", "desc"),
                ("quorum_factor", "1.0"),
            ],
        )

        _json = json.loads(await r.text())
        body = _json["message"]["body"]
        tracks = body["track_list"]

        tracks = sorted([(track, token_sort_ratio(search_term, track["track"]["track_name"])) for track in tracks], key=lambda x: x[1], reverse=True)
        if not tracks:
            return None
        return (await self.get_lrc_by_id(tracks[0][0]["track"]["track_id"]), tracks[0][1], 10000)



"""
Search for an LRC format (synchronized lyrics) of a music.

```py
import syncedlyrics
lrc_text = syncedlyrics.search("[TRACK_NAME] [ARTIST_NAME]")
```
"""

import asyncio
import aiohttp
from typing import Optional, List
import logging
from .providers import NetEase, Megalobiz, Musixmatch, Lrclib, Tencent
from .utils import is_lrc_valid, save_lrc_file

logger = logging.getLogger(__name__)


async def get_lrc_from_provider(provider, search_term, duration, allow_plain_format, max_deviation):
    logger.debug(f"Looking for an LRC on {provider.__class__.__name__}")
    result = await provider.get_lrc(search_term, duration=duration, max_deviation=max_deviation)

    if not result:
        return None
    lrc = result[0]
    if is_lrc_valid(lrc, allow_plain_format):
        logger.info(
            f'synced-lyrics found for "{search_term}" on {provider.__class__.__name__}'
        )
    return result


async def search(
    search_term: str,
    allow_plain_format: bool = False,
    save_path: str = None,
    providers: List[str] = None,
    duration: int = -1,
    max_deviation: int = 2000
) -> Optional[str]:
    """
    Returns the synced lyrics of the song in [LRC](https://en.wikipedia.org/wiki/LRC_(file_format)) format if found.
    ### Arguments
    - `search_term`: The search term to find the track
    - `allow_normal_format`: Return a plain text (not synced) lyrics if not LRC was found
    - `save_path`: Path to save `.lrc` lyrics. No saving if `None`
    - `providers`: A list of provider names to include in searching; loops over all the providers as soon as an LRC is found
    - `duration`: The duration of track in ms. Set below 0 if unknow
    - `max_deviation`: Max deviation for a subtitle length in ms, enable if duration is positive
    """
    session = aiohttp.ClientSession()
    _providers = [Musixmatch(session), NetEase(session), Megalobiz(session), Lrclib(session), Tencent(session)]
    if providers:
        # Filtering the providers
        _providers = [
            p for p in _providers if p.__class__.__name__ in providers]
    lrc = None
    tasks = [
        get_lrc_from_provider(provider, search_term, duration, allow_plain_format, max_deviation) for provider in _providers
    ]

    results = await asyncio.gather(*tasks)
    await session.close()

    results = [_ for _ in results if _]
    results = [_ for _ in results if _[0]]

    if results:
        if duration >= 0:
            lrc = sorted(results, key=lambda x: (x[1], -x[2]), reverse=True)[0][0]
        else:
            lrc = sorted(results, key=lambda x: (x[1],), reverse=True)[0][0]

    if not results:
        logger.info(f'No synced-lyrics found for "{search_term}" :(')
        return
    if save_path:
        save_path = save_path.format(search_term=search_term)
        save_lrc_file(save_path, lrc)
    return lrc



"""Megalobiz (megalobiz.com) LRC provider"""

from typing import Optional, Tuple
import re
import rapidfuzz
from bs4 import SoupStrainer
from .base import LRCProvider
from ..utils import generate_bs4_soup


class Megalobiz(LRCProvider):
    """Megabolz provider class"""

    ROOT_URL = "https://www.megalobiz.com"
    SEARCH_ENDPOINT = ROOT_URL + "/search/all?qry={q}&searchButton.x=0&searchButton.y=0"

    async def get_lrc(self, search_term: str, duration: int = -1, max_deviation: int = 2000) -> Tuple[Optional[str], int, int]:
        url = self.SEARCH_ENDPOINT.format(q=search_term)

        def href_match(h: Optional[str]):
            if h and h.startswith("/lrc/maker/"):
                return True
            return False

        def duration_match(s: str):
            _result = re.findall(r"\d{1,2}:\d{1,2}\.\d{1,2}", s)
            if _result:
                _time = _result[0].split(':')
                _time = list(map(float, _time))
                _duration = _time[0] * 60 * 1000 + _time[1] * 1000
                return _duration
            else:
                return

        a_tags_boud = SoupStrainer("a", href=href_match)
        soup = await generate_bs4_soup(self.session, url, parse_only=a_tags_boud)
        a_tag = soup.find_all("a")
        if not a_tag:
            return None

        reslut = []
        if duration >= 0:
            for tag in a_tag:
                ms = duration_match(tag["title"])
                if not ms:
                    continue
                offset = abs(duration - ms)
                if offset <= max_deviation:
                    reslut.append((tag, rapidfuzz.fuzz.token_sort_ratio(search_term, tag["title"]), offset,))
        else:
            reslut = [(tag, rapidfuzz.fuzz.token_sort_ratio(search_term, tag["title"]), 0) for tag in a_tag]

        if not reslut:
            return None
        reslut = sorted(reslut, key=lambda x: (x[1], -x[2], ), reverse=True)[0]

        # Scraping from the LRC page
        lrc_id = reslut[0]["href"].split(".")[-1]
        soup = await generate_bs4_soup(self.session, self.ROOT_URL + reslut[0]["href"])
        return (soup.find("div", {"id": f"lrc_{lrc_id}_details"}).get_text(), reslut[1], reslut[2])



"""Tencent (y.qq.com) china-based provider"""

import base64
import html
from typing import Optional, Tuple
import json
import urllib
from rapidfuzz.fuzz import token_sort_ratio
from .base import LRCProvider
from ..utils import format_lyrics

headers = {
    "Referer": "'https://y.qq.com'",
}


class Tencent(LRCProvider):
    """Tencent provider class"""

    API_ENDPOINT_METADATA = "https://shc6.y.qq.com/soso/fcgi-bin/search_for_qq_cp"
    API_ENDPOINT_LYRICS = "https://c.y.qq.com/lyric/fcgi-bin/fcg_query_lyric_new.fcg"

    def __init__(self, session) -> None:
        super().__init__(session)
        self.session.headers.update(headers)

    async def search_track(self, search_term: str, duration: int = -1, max_deviation: int = 2000) -> Tuple[Optional[dict], int, int]:
        params = {
            "format": 'json',
            "n": 20,
            "p": 1,
            "w": search_term,
            "cr": 1,
            "g_tk": 5381,
            "t": 0
        }
        response = await self.session.get(self.API_ENDPOINT_METADATA, params=params)
        text = await response.text()
        _json = json.loads(text)
        results = _json.get("data", {}).get("song")['list']
        if not results:
            return

        target = []
        if duration >= 0:
            for song in results:
                offset = abs(duration - song["interval"] * 1000)
                if offset <= max_deviation:
                    target.append((song, token_sort_ratio(search_term, song["songname"]), offset))
        else:
            target = [(song, token_sort_ratio(search_term, song["songname"]), 0) for song in results]

        if not target:
            return None

        target = sorted(target, key=lambda x: (x[1], -x[2], ), reverse=True)[0]

        return target

    async def get_lrc_by_id(self, track_id: str) -> Optional[str]:
        params = {
            "songmid": track_id,

            "g_tk": 5381,
            "loginUin": 0,
            "hostUin": 0,
            "inCharset": 'utf8',
            "outCharset": 'utf-8',
            "notice": 0,
            "platform": 'yqq',
            "needNewCode": 0,
            "format": "json"
        }
        response = await self.session.get(self.API_ENDPOINT_LYRICS, params=params)
        text = await response.text()
        _json = json.loads(text)

        lrc = _json.get("lyric", '')
        decoded_uri = urllib.parse.unquote(lrc)
        lrc = base64.b64decode(decoded_uri).decode("utf-8")
        lrc = html.unescape(lrc)

        tlyric = _json.get("trans", '')
        decoded_uri = urllib.parse.unquote(tlyric)
        tlyric = base64.b64decode(decoded_uri).decode("utf-8")
        tlyric = html.unescape(tlyric)

        lrc = format_lyrics(lrc, tlyric)
        if not lrc:
            return
        return lrc

    async def get_lrc(self, search_term: str, duration: int = -1, max_deviation: int = 2000) -> Tuple[Optional[str], int, int]:
        track = await self.search_track(search_term, duration, max_deviation)
        if not track:
            return
        return (await self.get_lrc_by_id(track[0]["songmid"]), track[1], track[2])



"""Deezer LRC provider (API powered by LyricFind)"""

from typing import Optional
from .base import LRCProvider

# Currently broken
# TODO: Fix invalid CSRF token

# Mostly based on https://gist.github.com/akashrchandran/95915c2081815397c454bd8aa4a118b5


class Deezer(LRCProvider):
    """Deezer provider class"""

    SEARCH_ENDPOINT = "https://api.deezer.com/search?q="
    API_ENDPOINT = "http://www.deezer.com/ajax/gw-light.php"
    token = "null"

    def __init__(self) -> None:
        super().__init__()
        self.token = self._api_call("deezer.getUserData")["results"]["checkForm"]

    def _api_call(self, method: str, json=None) -> dict:
        params = {
            "api_version": "1.0",
            "api_token": self.token,
            "input": "3",
            "method": method,
        }
        response = self.session.post(self.API_ENDPOINT, params=params, json=json)
        return response.json()

    def get_lrc_by_id(self, track_id: str) -> Optional[str]:
        lrc_response = self._api_call("song.getLyrics", json={"sng_id": track_id})
        lrc_json_objs = lrc_response["results"].get("LYRICS_SYNC_JSON")
        if not lrc_json_objs:
            # Returning the plain text lyrics
            return lrc_response["results"].get("LYRICS_TEXT")
        lrc = ""
        for chunk in lrc_json_objs:
            if chunk.get("lrc_timestamp") and chunk.get("line"):
                lrc += f"{chunk['lrc_timestamp']} {chunk['line']}\n"
        return lrc or None

    def get_lrc(self, search_term: str) -> Optional[str]:
        search_results = self.session.get(self.SEARCH_ENDPOINT + search_term).json()
        if not search_results.get("data"):
            return
        song = search_results.get("data")[0]
        return self.get_lrc_by_id(song["id"])



"""Lrclib (lrclib.net) LRC provider"""

from typing import Optional, Tuple
from rapidfuzz.fuzz import token_sort_ratio
from .base import LRCProvider


class Lrclib(LRCProvider):
    """Lrclib LRC provider class"""

    ROOT_URL = "https://lrclib.net"
    API_ENDPOINT = ROOT_URL + "/api"
    SEARCH_ENDPOINT = API_ENDPOINT + "/search"
    LRC_ENDPOINT = API_ENDPOINT + "/get/"

    async def get_lrc_by_id(self, track_id: str) -> Optional[str]:
        url = self.LRC_ENDPOINT + track_id
        r = await self.session.get(url)
        if not r.ok:
            return
        track = await r.json()
        return track.get("syncedLyrics", track.get("plainLyrics"))

    async def get_lrc(self, search_term: str, duration: int = -1, max_deviation: int = 2000) -> Tuple[Optional[str], int]:
        """Duration NotImplemented"""
        url = self.SEARCH_ENDPOINT
        r = await self.session.get(url, params={"q": search_term})
        if not r.ok or not r.content:
            print("bad")
            return
        tracks = await r.json()
        if not tracks:
            return
        tracks = sorted([(track, token_sort_ratio(search_term, track["name"])) for track in tracks], key=lambda x: x[1], reverse=True)

        _id = str(tracks[0][0]["id"])
        return (await self.get_lrc_by_id(_id), tracks[0][1], 10000)



"""NetEase (music.163.com) china-based provider"""

from typing import Optional, Tuple
import json
from rapidfuzz.fuzz import token_sort_ratio
from .base import LRCProvider
from ..utils import format_lyrics

headers = {
    "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
    "accept-language": "en-US,en;q=0.9,fa;q=0.8",
    "cache-control": "max-age=0",
    "sec-ch-ua": '".Not/A)Brand";v="99", "Google Chrome";v="103", "Chromium";v="103"',
    "sec-ch-ua-mobile": "?0",
    "sec-ch-ua-platform": '"Windows"',
    "sec-fetch-dest": "document",
    "sec-fetch-mode": "navigate",
    "sec-fetch-site": "none",
    "sec-fetch-user": "?1",
    "upgrade-insecure-requests": "1",
    "cookie": "NMTID=00OAVK3xqDG726ITU6jopU6jF2yMk0AAAGCO8l1BA; JSESSIONID-WYYY=8KQo11YK2GZP45RMlz8Kn80vHZ9%2FGvwzRKQXXy0iQoFKycWdBlQjbfT0MJrFa6hwRfmpfBYKeHliUPH287JC3hNW99WQjrh9b9RmKT%2Fg1Exc2VwHZcsqi7ITxQgfEiee50po28x5xTTZXKoP%2FRMctN2jpDeg57kdZrXz%2FD%2FWghb%5C4DuZ%3A1659124633932; _iuqxldmzr_=32; _ntes_nnid=0db6667097883aa9596ecfe7f188c3ec,1659122833973; _ntes_nuid=0db6667097883aa9596ecfe7f188c3ec; WNMCID=xygast.1659122837568.01.0; WEVNSM=1.0.0; WM_NI=CwbjWAFbcIzPX3dsLP%2F52VB%2Bxr572gmqAYwvN9KU5X5f1nRzBYl0SNf%2BV9FTmmYZy%2FoJLADaZS0Q8TrKfNSBNOt0HLB8rRJh9DsvMOT7%2BCGCQLbvlWAcJBJeXb1P8yZ3RHA%3D; WM_NIKE=9ca17ae2e6ffcda170e2e6ee90c65b85ae87b9aa5483ef8ab3d14a939e9a83c459959caeadce47e991fbaee82af0fea7c3b92a81a9ae8bd64b86beadaaf95c9cedac94cf5cedebfeb7c121bcaefbd8b16dafaf8fbaf67e8ee785b6b854f7baff8fd1728287a4d1d246a6f59adac560afb397bbfc25ad9684a2c76b9a8d00b2bb60b295aaafd24a8e91bcd1cb4882e8beb3c964fb9cbd97d04598e9e5a4c6499394ae97ef5d83bd86a3c96f9cbeffb1bb739aed9ea9c437e2a3; WM_TID=AAkRFnl03RdABEBEQFOBWHCPOeMra4IL; playerid=94262567",
}


class NetEase(LRCProvider):
    """NetEase provider class"""

    API_ENDPOINT_METADATA = "https://music.163.com/api/search/pc"
    API_ENDPOINT_LYRICS = "https://music.163.com/api/song/lyric"

    def __init__(self, session) -> None:
        super().__init__(session)
        self.session.headers.update(headers)

    async def search_track(self, search_term: str, duration: int = -1, max_deviation: int = 2000) -> Tuple[Optional[dict], int, int]:
        """Returns a `dict` containing some metadata for the found track."""
        params = {"limit": 20, "type": 1, "offset": 0, "s": search_term}
        response = await self.session.get(self.API_ENDPOINT_METADATA, params=params)
        text = await response.text()
        _json = json.loads(text)
        results = _json.get("result", {}).get("songs")
        if not results:
            return

        target = []
        if duration >= 0:
            for song in results:
                offset = abs(duration - song["duration"])
                if offset <= max_deviation:
                    target.append([song, token_sort_ratio(search_term, song["name"]), offset])
        else:
            target = [(song, token_sort_ratio(search_term, song["name"]), 0) for song in results]

        if not target:
            return None

        target = sorted(target, key=lambda x: (x[1], -x[2], ), reverse=True)[0]

        return target

    async def get_lrc_by_id(self, track_id: str) -> Optional[str]:
        params = {"id": track_id, "lv": -1, "tv": -1}
        response = await self.session.get(self.API_ENDPOINT_LYRICS, params=params)
        text = await response.text()
        _json = json.loads(text)
        lrc = _json.get("lrc", {}).get("lyric") or ''
        tlyric = _json.get("tlyric", {}).get("lyric") or ''
        lrc = format_lyrics(lrc, tlyric)
        if not lrc:
            return
        return lrc

    async def get_lrc(self, search_term: str, duration: int = -1, max_deviation: int = 2000) -> Tuple[Optional[str], int, int]:
        track = await self.search_track(search_term, duration, max_deviation)
        if not track:
            return
        return (await self.get_lrc_by_id(track[0]["id"]), track[1], track[2])



"""Synchronized Lyrics providers"""

from .netease import NetEase
from .deezer import Deezer
from .lyricsify import Lyricsify
from .megalobiz import Megalobiz
from .musixmatch import Musixmatch
from .lrclib import Lrclib
from .tencent import Tencent



"""Lyricsify (lyricsify.com) LRC provider"""

from typing import Optional
from bs4 import SoupStrainer
import rapidfuzz
from .base import LRCProvider
from ..utils import generate_bs4_soup

# Currently broken
# TODO: Bypassing Cloudflare anti-bot system


class Lyricsify(LRCProvider):
    """Lyricsify LRC provider class"""

    ROOT_URL = "https://www.lyricsify.com"
    SEARCH_ENDPOINT = ROOT_URL + "/search?q="

    def __init__(self) -> None:
        super().__init__()
        self.parser = "html.parser"

    def get_lrc(self, search_term: str) -> Optional[str]:
        url = self.SEARCH_ENDPOINT + search_term.replace(" ", "+")

        # Just processing the `a` tags whose `href` attribute starts with /lyric/
        # and whose text is similar to the query too. https://github.com/maxbachmann/RapidFuzz#scorers
        _t = lambda s: s.lower().replace("-", "") if s else s
        text_match = lambda t: rapidfuzz.fuzz.token_sort_ratio(_t(search_term), _t(t))
        href_match = lambda h: h.startswith("/lyric/")
        a_tags_boud = SoupStrainer("a", href=href_match)
        soup = generate_bs4_soup(self.session, url, parse_only=a_tags_boud)
        a_tag = soup.find_all("a", string=lambda t: text_match(t) > 80, limit=4)
        if not a_tag:
            return None

        # The one with the highest similarity score
        a_tag = sorted(a_tag, key=lambda tag: text_match(tag.string))[-1]

        # Scraping from the LRC page
        lrc_id = a_tag["href"].split(".")[-1]
        soup = generate_bs4_soup(self.session, self.ROOT_URL + a_tag["href"])
        return soup.find("div", {"id": f"lyrics_{lrc_id}_details"}).get_text()



"""Spotify LRC provider (the API is powered by Musixmatch)"""

from typing import Optional
from .base import LRCProvider


class Spotify(LRCProvider):
    """Spotify provider class"""

    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def get_track_id(search_term: str) -> Optional[str]:
        """Returns a Spotify track ID for given `search_term`"""
        # TODO: self.client.search(search_term) and processing the results
        raise NotImplementedError

    def get_lrc_by_id(self, track_id: str) -> Optional[str]:
        # TODO:
        raise NotImplementedError

    def get_lrc(self, search_term: str) -> Optional[str]:
        # TODO: Use https://github.com/akashrchandran/spotify-lyrics-api
        raise NotImplementedError



"""Utility functions for `syncedlyrics` package"""

import re
from bs4 import BeautifulSoup, FeatureNotFound


def is_lrc_valid(lrc: str, allow_plain_format: bool = False) -> bool:
    """Checks whether a given LRC string is valid or not."""
    if not lrc:
        return False
    if not allow_plain_format:
        if not ("[" in lrc and "]" in lrc):
            return False

    return True


def save_lrc_file(path: str, lrc_text: str):
    """Saves the `.lrc` file"""
    with open(path, "w", encoding="utf-8") as f:
        f.write(lrc_text)


async def generate_bs4_soup(session, url: str, **kwargs):
    """Returns a `BeautifulSoup` from the given `url`.
    Tries to use `lxml` as the parser if available, otherwise `html.parser`
    """
    r = await session.get(url)
    text = await r.text()
    try:
        soup = BeautifulSoup(text, features="lxml", **kwargs)
    except FeatureNotFound:
        soup = BeautifulSoup(text, features="html.parser", **kwargs)
    return soup


def trim_lyric(lyric):
    result = []
    lines = lyric.split('\n')
    for line in lines:
        match = re.match(r'^\[(\d{2}):(\d{2}\.\d*)\](.*)$', line)
        if match:
            result.append({
                'time': int(int(match[1]) * 60 * 1000 + float(match[2]) * 1000),
                'text': match[3]
            })
    return sorted(result, key=lambda x: x['time'])


def format_lyrics(lyric, tlyric):
    lyric_array = trim_lyric(lyric)
    tlyric_array = trim_lyric(tlyric)

    if len(tlyric_array) == 0:
        return lyric

    result = []
    j = 0

    for i in range(len(lyric_array)):
        if j == len(tlyric_array):
            break

        time = lyric_array[i]['time']
        text = lyric_array[i]['text']

        while time > tlyric_array[j]['time'] and j + 1 < len(tlyric_array):
            j += 1

        if time == tlyric_array[j]['time'] and len(tlyric_array[j]['text']) > 0:
            text = f"{text} ({tlyric_array[j]['text']})"

        result.append({
            'time': time,
            'text': text
        })

    formatted_result = []

    for x in result:
        minus = str(x['time'] // 60000).zfill(2)
        second = str((x['time'] % 60000) // 1000).zfill(2)
        millisecond = str(x['time'] % 1000).zfill(3)

        formatted_result.append(f"[{minus}:{second}.{millisecond}]{x['text']}")

    return '\n'.join(formatted_result)
